package org.objectweb.telosys.screen.env;

import javax.servlet.http.HttpServletRequest;

import org.objectweb.telosys.common.Telosys;
import org.objectweb.telosys.common.TelosysClassLogger;
import org.objectweb.telosys.common.TelosysException;
import org.objectweb.telosys.common.TelosysFactory;
import org.objectweb.telosys.rpl.ScreenRequestParameters;
import org.objectweb.telosys.screen.core.ScreenContext;

/**
 * Set of static methods for Screen Context management
 * 
 * @author Laurent GUERIN
 *  
 */
public final class ScreenContextManager
{
    private static final TelosysClassLogger log = new TelosysClassLogger(ScreenContextManager.class);
    
    private static final ScreenRequestParameters VOID_PARAMS = new ScreenRequestParameters();
    
    /**
     * No constructor for this class
     */
    private ScreenContextManager()
    {
    }

    //---------------------------------------------------------------------------
    /**
     * Try to find a ScreenContext if possible <br>
     * Return NULL if cannot be found ( no error ) 
     * @param httpRequest
     * @param iScreenId
     * @return the ScreenContext found or null if not found
     */
    public static ScreenContext findScreenContext(HttpServletRequest httpRequest, int iScreenId) 
    {
        if (httpRequest == null)
        {
            return null ;
        }
        //--- Get the screen session from the Http Request
        ScreenSession screenSession = ScreenSessionManager.getScreenSession(httpRequest);
        if (screenSession == null)
        {
            return null ;
        }
        //--- Get the screen context from the screen session
        return screenSession.getScreenContext(iScreenId);        
    }

    //---------------------------------------------------------------------------
    /**
     * Try to find a ScreenContext if possible <br>
     * Return NULL if cannot be found ( no error ) 
     * @param httpRequest
     * @param iScreenId
     * @param sScreenName
     * @return the ScreenContext found or null if not found
     */
    public static ScreenContext findScreenContext(HttpServletRequest httpRequest, int iScreenId, String sScreenName) 
    {
		if ( sScreenName == null )
		{
            return null ;
		}
		if ( sScreenName.trim().length() == 0 )
		{
            return null ;
		}
		
        //--- Find the screen context
		ScreenContext screenContext = ScreenContextManager.findScreenContext( httpRequest, iScreenId );
		if (screenContext != null) 
		{
            //--- Is it the expected screen name ? 
            if ( sScreenName.trim().equalsIgnoreCase( screenContext.getScreenName() ) )
            {
                return screenContext ;                
            }
		}
		return null ;        
    }
    
    //---------------------------------------------------------------------------
    // GET ( TRY TO RETRIEVE AN EXISTING SCREEN CONTEXT )
    //---------------------------------------------------------------------------
    /**
     * Try to retrieve an existing screen context associated with the given screen ID
     * @param request
     * @param iScreenId
     * @return the ScreenContext or null if not found 
     * @throws TelosysException
     */
    public static ScreenContext getScreenContext(HttpServletRequest request, int iScreenId) throws TelosysException
    {
        if (request == null)
        {
            throw new TelosysException("getScreenContext : request is null !");
        }

        //--- Get the screen session from the Http Request
        ScreenSession screenSession = ScreenSessionManager.getScreenSession(request);
        if (screenSession != null)
        {
            //--- Get the screen context from the screen session
            return screenSession.getScreenContext(iScreenId);        
        }
        else
        {
            throw new TelosysException("getScreenContext : cannot get Screen Session !");            
        }        
    }
    
    /**
     * Try to retrieve an existing screen context associated with the given Screen ID and Screen Name 
     * @param httpRequest
     * @param iScreenId
     * @param sScreenName
     * @return the ScreenContext or null if not found 
     * @throws TelosysException
     */
    public static ScreenContext getScreenContextForName( HttpServletRequest httpRequest, int iScreenId, String sScreenName ) throws TelosysException
    {
        if (httpRequest == null)
        {
            throw new TelosysException("getScreenContextForName : request is null !");
        }
		if ( iScreenId < 0 )
		{
            throw new TelosysException("getScreenContextForName : invalid Screen Id " + iScreenId );
		}
		if ( sScreenName == null )
		{
            throw new TelosysException("getScreenContextForName : screen name is null !");
		}
		if ( sScreenName.trim().length() == 0 )
		{
            throw new TelosysException("getScreenContextForName : screen name is void !");
		}
		
        //--- Get the screen context
		ScreenContext screenContext = ScreenContextManager.getScreenContext( httpRequest, iScreenId );
		if (screenContext != null) 
		{
            //--- Is it the expected screen name ? 
            if ( sScreenName.equalsIgnoreCase( screenContext.getScreenName() ) )
            {
                return screenContext ;                
            }
		}
		return null ;
    }
    
    //---------------------------------------------------------------------------
    // OPEN
    //---------------------------------------------------------------------------
    /**
     * Open (create) a new Screen Context 
     * @param request
     * @param sScreenName
     * @param iScreenId the ScreenContext ID to use, or (since 1.0.4) a negative value for "next available ID"
     * @return
     * @throws TelosysException
     */
    public static ScreenContext openScreenContext(HttpServletRequest request, String sScreenName, int iScreenId ) throws TelosysException
    {
        log.trace("openScreenContext(HttpServletRequest," + sScreenName +"," + iScreenId + ")...");
        return openScreenContext(request, sScreenName, iScreenId, null);
    }
    
    //---------------------------------------------------------------------------
    /**
     * Open (create) a new Screen Context with parameters
     * @param request
     * @param sScreenName
     * @param iScreenId the ScreenContext ID to use, or (since 1.0.4) a negative value for "next available ID"
     * @param params
     * @return
     * @throws TelosysException
     */
    public static ScreenContext openScreenContext(HttpServletRequest request, String sScreenName, int iScreenId,
            ScreenRequestParameters params ) throws TelosysException
    {
        log.trace("openScreenContext(HttpServletRequest," + sScreenName +"," + iScreenId + "," + (params!=null?"params":"null") + ")...");
        if (request == null)
        {
            throw new TelosysException("openScreenContext : request is null !");
        }

        //--- Get the screen session
        ScreenSession screenSession = ScreenSessionManager.getScreenSession(request);
        if (screenSession == null)
        {
            throw new TelosysException("openScreenContext : cannot get Screen Session !");
        }
        
        return openScreenContext(screenSession, sScreenName, iScreenId, params);
    }

    //---------------------------------------------------------------------------
    /**
     * Open (create) a new Screen Context
     * @param screenSession
     * @param sScreenName
     * @param iScreenId the ScreenContext ID to use, or (since 1.0.4) a negative value for "next available ID"
     * @return
     * @throws TelosysException
     */
    public static ScreenContext openScreenContext(ScreenSession screenSession, String sScreenName, int iScreenId ) throws TelosysException
    {
        log.trace("openScreenContext(ScreenSession," + sScreenName +"," + iScreenId + " )...");
        return openScreenContext(screenSession, sScreenName, iScreenId, null);
    }
    
    //---------------------------------------------------------------------------
    /**
     * Open (create) a new Screen Context with parameters
     * @param screenSession
     * @param sScreenName the name of the ScreenContext
     * @param iScreenId the ScreenContext ID to use, or (since 1.0.4) a negative value for "next available ID"
     * @param params 
     * @return
     * @throws TelosysException
     */
    public static ScreenContext openScreenContext(ScreenSession screenSession, String sScreenName, int iScreenId,
            ScreenRequestParameters params ) throws TelosysException
    {
        log.trace("openScreenContext(ScreenSession," + sScreenName +"," + iScreenId + "," + (params!=null?"params":"null") + " )...");
        if (screenSession == null)
        {
            throw new TelosysException("openScreenContext : screen session is null !");
        }

//---- new in v 1.0.4 ----------
//        //--- Get the screen factory
//        TelosysFactory factory = Telosys.getFactory();
//        if (factory == null)
//        {
//            throw new TelosysException("openScreenContext : Cannot get the TelosysFactory !");
//        }
//
//        //--- Create the new screen context
//        ScreenContext screenContext = factory.createScreen(screenSession, sScreenName, iScreenId );
//        if (screenContext == null)
//        {
//            throw new TelosysException("openScreenContext : Cannot create the screen (factory returns null) !");
//        }
//
//        //--- Call the "create" event in the ScreenManager if any 
//        screenContext.eventCreate( screenSession, ( params != null ? params : VOID_PARAMS ) );
//        
//        //--- Set the new screen in the current session
//        screenSession.setScreenContext(iScreenId, screenContext);

        //--- Create the new ScreenContext and store it in the ScreenSession
        ScreenContext screenContext = null ;
        if ( iScreenId < 0 )
        {
            //--- Undefined ScreenContext ID => try to use the next available ScreenContext
            screenContext = openNextAvailableScreenContext(screenSession, sScreenName );
        }
        else
        {
            //--- Use the given ScreenContext ID 
            if (iScreenId > ScreenSession.getMaxScreenId())
            {
                throw new TelosysException("openScreenContext : Invalid screen Id ( " + iScreenId + " > " + ScreenSession.getMaxScreenId() + " ) ! ");
            }
            screenContext = openScreenContextWithId(screenSession, sScreenName, iScreenId); 
        }

		//--- Call the "create" event in the ScreenManager if any 
		screenContext.eventCreate( screenSession, ( params != null ? params : VOID_PARAMS ) );        

//---- new in v 1.0.4 ----------

		return screenContext;
    }
    
    //---------------------------------------------------------------------------
    /**
     * Open a new ScreenContext with the given ScreenContext ID
     * @param screenSession
     * @param sScreenName
     * @param iScreenContextId
     * @return
     * @throws TelosysException
     * @since 1.0.4
     */
    private static ScreenContext openScreenContextWithId(ScreenSession screenSession, String sScreenName, int iScreenContextId) 
    		throws TelosysException
    {
        //--- The iScreenContextId validity has been checked before (in the RPL )

        //--- Create the new screen context with the given ScreenContext ID
        ScreenContext screenContext = createScreenContext(screenSession, sScreenName, iScreenContextId );

        //--- Set the new screen in the current session
        screenSession.setScreenContext(iScreenContextId, screenContext);

        return screenContext;
    }
    
    //---------------------------------------------------------------------------
    /**
     * Open a new ScreenContext using the next available ScreenContext ID 
     * @param screenSession
     * @param sScreenName
     * @return
     * @throws TelosysException
     * @since 1.0.4
     */
    private static ScreenContext openNextAvailableScreenContext(ScreenSession screenSession, String sScreenName )
    		throws TelosysException
    {
        log.trace("openNextAvailableScreenContext(ScreenSession," + sScreenName + " )...");
        if (screenSession == null)
        {
            throw new TelosysException("openScreenContext : screen session is null !");
        }

        //--- Try to find and reserve an available slot in the ScreenSession 
        int iScreenContextId = screenSession.getAvailableScreenContextId();
        if ( iScreenContextId >= 0 )
        {
            //--- Slot available
            try
            {
                //--- Create the new screen context with the new ScreenContext ID
                ScreenContext screenContext = createScreenContext(screenSession, sScreenName, iScreenContextId );
                //--- Store the new screen context in the session ( in the reserved slot )
                screenSession.setScreenContext(iScreenContextId, screenContext);
                return screenContext ;
                
            } catch (TelosysException e) // Catch all Exceptions to be sure to free the reserved slot if an error occurs
            {
                screenSession.freeScreenContextSlot(iScreenContextId);
                throw e ;
            } catch (Throwable t) 
            {
                screenSession.freeScreenContextSlot(iScreenContextId);
                throw new TelosysException("Unexpected Exception", t) ;
            }
        }
        else
        {
            //--- NO Slot available !
            throw new TelosysException("openScreenContext : No slot available for a new screen context !");
        }
    }
    
    //---------------------------------------------------------------------------
    /**
     * Create a new instance of ScreenContext
     * @param screenSession
     * @param sScreenName
     * @param iScreenId
     * @return
     * @throws TelosysException
     * @since 1.0.4
     */
    private static ScreenContext createScreenContext(ScreenSession screenSession, String sScreenName, int iScreenId) throws TelosysException
    {
        log.trace("createScreenContext(ScreenSession," + sScreenName +"," + iScreenId + " )...");

        //--- Get the screen factory
        TelosysFactory factory = Telosys.getFactory();
        if (factory == null)
        {
            throw new TelosysException("createScreenContext : Cannot get the TelosysFactory !");
        }

        //--- Create the new screen context
        ScreenContext screenContext = factory.createScreen(screenSession, sScreenName, iScreenId );
        if (screenContext == null)
        {
            throw new TelosysException("createScreenContext : Cannot create the screen (factory returns null) !");
        }
        return screenContext ;
    }
    
    //---------------------------------------------------------------------------
    // CLOSE
    //---------------------------------------------------------------------------
    /**
     * Close screen context
     * @param request
     * @param iScreenId
     * @throws TelosysException
     */
    public static void closeScreenContext(HttpServletRequest request, int iScreenId) throws TelosysException
    {
        log.trace("closeScreenContext(HttpServletRequest," + iScreenId + ")...");
        if (request == null)
        {
            throw new TelosysException("closeScreenContext : request is null !");
        }

        //--- Get the screen session
        //ScreenSession screenSession = ScreenSessionManager.getScreenSession(request);
        ScreenSession screenSession = ScreenSessionManager.findScreenSession(request);
        if (screenSession == null)
        {
            throw new TelosysException("closeScreenContext : cannot get Screen Session !");
        }
        
        closeScreenContext(screenSession, iScreenId);
    }
    
    //---------------------------------------------------------------------------
    /**
     * Close screen context
     * @param screenSession
     * @param iScreenId
     * @throws TelosysException
     */
    public static void closeScreenContext(ScreenSession screenSession, int iScreenId) throws TelosysException
    {
        log.trace("closeScreenContext(ScreenSession," + iScreenId + ")...");
        if (screenSession == null)
        {
            throw new TelosysException("closeScreenContext : screen session is null !");
        }

        //--- Remove the screen from the screen session
        screenSession.removeScreenContext(iScreenId);
    }
    
    //---------------------------------------------------------------------------
    // USE  ( RETRIEVE AN EXISTING SCREEN or CREATE A NEW ONE IF IT DOESN'T EXIST )
    //---------------------------------------------------------------------------
    /**
     * Retrieve an existing screen context or create a new one if it doesn't exist
     * @param request
     * @param sScreenName
     * @param iScreenId
     * @return
     * @throws TelosysException
     */
    public static ScreenContext useScreenContext(HttpServletRequest request, String sScreenName, int iScreenId ) throws TelosysException
    {
        log.trace("useScreenContext(HttpServletRequest," + sScreenName +"," + iScreenId + ")...");
        return useScreenContext(request, sScreenName, iScreenId, null);
    }

    //---------------------------------------------------------------------------
    /**
     * Retrieve an existing screen context or create a new one if it doesn't exist
     * @param request
     * @param sScreenName
     * @param iScreenId
     * @param params
     * @return
     * @throws TelosysException
     */
    public static ScreenContext useScreenContext(HttpServletRequest request, String sScreenName, int iScreenId,
            ScreenRequestParameters params) throws TelosysException
    {
        log.trace("useScreenContext(HttpServletRequest," + sScreenName +"," + iScreenId + "," + (params!=null?"params":"null") + ")...");
        if (request == null)
        {
            throw new TelosysException("useScreenContext : request is null !");
        }

        //--- Get the screen session
        ScreenSession screenSession = ScreenSessionManager.getScreenSession(request);
        if (screenSession == null)
        {
            throw new TelosysException("useScreenContext : cannot get Screen Session !");
        }
        return useScreenContext(screenSession, sScreenName, iScreenId, params );
    }
    
    //---------------------------------------------------------------------------
    /**
     * Retrieve an existing screen context or create a new one if it doesn't exist
     * @param screenSession
     * @param sScreenName
     * @param iScreenId
     * @return
     * @throws TelosysException
     */
    public static ScreenContext useScreenContext(ScreenSession screenSession, String sScreenName, int iScreenId ) throws TelosysException
    {
        log.trace("useScreenContext(ScreenSession," + sScreenName +"," + iScreenId + ")...");
        return useScreenContext(screenSession, sScreenName, iScreenId, null);
    }
    
    //---------------------------------------------------------------------------
    /**
     * Retrieve an existing screen context or create a new one if it doesn't exist
     * @param screenSession
     * @param sScreenName
     * @param iScreenId the ScreenContext ID ( cannot be < 0 )
     * @param params
     * @return
     * @throws TelosysException
     */
    public static ScreenContext useScreenContext(ScreenSession screenSession, String sScreenName, int iScreenId,
            ScreenRequestParameters params) throws TelosysException
    {
        log.trace("useScreenContext(ScreenSession," + sScreenName +"," + iScreenId + "," + (params!=null?"params":"null") + ")...");
        if (screenSession == null)
        {
            throw new TelosysException("useScreenContext : screen session is null !");
        }

        if ( iScreenId < 0 ) // v 1.0.4
        {
            throw new TelosysException("useScreenContext : screen id = " + iScreenId + " ( must be > 0 for 'USE' action ) ");
        }


        //--- Get the screen context
        ScreenContext screenContext = screenSession.getScreenContext(iScreenId);
        if ( screenContext != null )
        {
            //--- Is it the expected screen name ? 
            if ( sScreenName.equalsIgnoreCase( screenContext.getScreenName() ) )
            {
                //--- Re-use the current screen context ...
                //--- Call the "reuse" event in the ScreenData manager if any 
                screenContext.eventReuse( screenSession, ( params != null ? params : VOID_PARAMS ) );
                
                return screenContext ;                
            }
            else
            {
                //--- It's not the expected screen => cannot use it => open the expected screen context 
                return openScreenContext(screenSession, sScreenName, iScreenId, params);
            }
        }
        else
        {
            return openScreenContext(screenSession, sScreenName, iScreenId, params);
        }
    }

}