package org.objectweb.telosys.uil.taglib.widget;

import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.Properties;

import javax.servlet.jsp.JspWriter;
import javax.servlet.jsp.PageContext;

import org.objectweb.telosys.common.Tool;
import org.objectweb.telosys.uil.taglib.TagCommons;

/**
 * @author T. Poingt / L. Guerin
 * 
 * This static class provides methods to managed the default values for
 * different tags. The default values are initialized from a standard properties
 * file using the syntax : tag.attribute = value
 */
public class TagDefaultValues
{

    private static Properties   _propDefaultValues           = null;

    /**
     * Init the default values with the given properties file name
     * @param sPropFileName
     * @return true if OK, else false
     */
    public static boolean init(String sPropFileName)
    {
        _propDefaultValues = Tool.loadProperties(sPropFileName);
        return ( _propDefaultValues != null ) ;
    }

    /**
     * Init the default values with the given properties InputStream
     * @param is
     * @return
     */
    public static boolean init( InputStream is )
    {
        _propDefaultValues = Tool.loadProperties(is);
        return ( _propDefaultValues != null ) ; 
    }

    /**
     * Return the default attribute value for a given tag name 
     * 
     * @param sTag
     * @param sAttrib
     * @return default value (or null if no default value)
     */
    private static String getGlobalDefaultAttributeValue(String sTag, String sAttrib)
    {
        if (_propDefaultValues != null)
        {
            return (String) _propDefaultValues.get(sTag + "." + sAttrib);
        }
        else
        {
            return null;
        }
    }

    /**
     * Return the default attribute value for the given tag name and attribute name
     * @param tagName
     * @param sAttribute
     * @return the default value if any, or null
     */
    public static String getDefaultAttributeValue(PageContext pageContext, String tagName, String sAttribute)
    {
        String sRetValue = null;
        
        //--- 1) Try to find a local default value ( set in the current page context ) 
        // A default value set with the "default" tag
        TagLocalDefaultValues localDefaults = TagCommons.getLocalDefaultValues(pageContext);
        if (localDefaults != null)
        {
            sRetValue = localDefaults.getDefaultAttributeValue(tagName, sAttribute);
            if (sRetValue == null)
            {
                // Try to get a default value for all tags
                sRetValue = localDefaults.getDefaultAttributeValue("*", sAttribute);
            }
        }

        //--- 2) If not found, try to get a global default value (same for all screens)
        if (sRetValue == null)
        {
            // The global default values are initialized from a properties file
            sRetValue = getGlobalDefaultAttributeValue(tagName, sAttribute);
            if (sRetValue == null)
            {
                // Try to get a default value for all tags
                sRetValue = getGlobalDefaultAttributeValue("*", sAttribute);
            }
        }
        return sRetValue;
    }
    
    public static void printDefaults(JspWriter out)
    {
        if (_propDefaultValues != null)
        {
	        String sKey = null ;
	        Enumeration keys = _propDefaultValues.keys();
	        try
	        {
	            out.println("Global defaults for tags attributes : ");
	            if ( keys != null )
	            {
	                while ( keys.hasMoreElements() )
	                {
	                    sKey = (String) keys.nextElement();
	                    out.println(" . '" + sKey + "' : '" + _propDefaultValues.get(sKey) + "'" );
	                }            
	            }
	            else
	            {
	                out.println("void");
	            }
	        } catch (IOException e)
	        {
	            // Nothing to do 
	        }
        }
        else
        {
	        try
	        {
	            out.println("No global defaults for tags attributes");
	        } catch (IOException e)
	        {
	            // Nothing to do 
	        }
        }
    }
}